"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.VirtualFileSystem = void 0;
const fs_extra_1 = __importDefault(require("fs-extra"));
const path_1 = __importDefault(require("path"));
const lodash_1 = require("lodash");
class VirtualFileSystem {
    constructor({ cwd }) {
        this.fileSystemAdapterMethods = {
            readdirSync: this.readdirSync.bind(this),
            lstatSync: this.lstatSync.bind(this),
            statSync: this.statSync.bind(this),
        };
        this.cwd = cwd;
        this.map = new Map();
    }
    getFile(path) {
        if (!this.map.has(path)) {
            this.map.set(path, { action: 'none' });
        }
        return this.map.get(path);
    }
    // @ts-ignore
    statSync(path, options) {
        if (options) {
            throw new Error('need to implenent bigIntOptions');
        }
        path = path.toString();
        if (!this.map.has(path)) {
            this.map.set(path, { action: 'none' });
        }
        const file = this.getFile(path);
        if (!file.stats) {
            const stats = fs_extra_1.default.statSync(path);
            this.map.set(path, { ...file, stats });
        }
        return this.getFile(path).stats;
    }
    // @ts-expect-error - options type is too big
    readdirSync(path, options) {
        const files = fs_extra_1.default.readdirSync(path, options);
        const writtenFilePaths = this.getWrittenFilePaths()
            .filter((p) => path_1.default.dirname(p) === path)
            .map((p) => new DirentLike(p));
        const removedFilePaths = this.getRemovedFilePaths().filter((p) => path_1.default.dirname(p) === path);
        const allFilePaths = [...files, ...writtenFilePaths];
        // @ts-expect-error
        const all = allFilePaths.filter((p) => !removedFilePaths.includes(p.name));
        // @ts-expect-error
        return all;
    }
    getWrittenFilePaths() {
        const written = [];
        this.map.forEach((fileInformation, path) => {
            if (fileInformation.action === 'write') {
                written.push(path);
            }
        });
        return written;
    }
    getRemovedFilePaths() {
        const removed = [];
        this.map.forEach((fileInformation, path) => {
            if (fileInformation.action === 'remove') {
                removed.push(path);
            }
        });
        return removed;
    }
    writeFileSync(path, data) {
        const file = this.getFile(path);
        file.data = data;
        file.action = 'write';
        // @ts-expect-error
        file.stats = new StatsLike();
    }
    removeSync(path) {
        const file = this.getFile(path);
        file.action = 'remove';
    }
    // @ts-expect-error - options type is too big
    lstatSync(path) {
        path = path.toString();
        if (!this.map.has(path)) {
            this.map.set(path, { action: 'none' });
        }
        const file = this.getFile(path);
        if (!file.stats) {
            const stats = fs_extra_1.default.lstatSync(path);
            this.map.set(path, { ...file, stats });
        }
        return this.getFile(path).stats;
    }
    readFileSync(path) {
        const file = this.getFile(path);
        if (file.action === 'remove') {
            // We'll might need to deal with that later,
            // because globby can read file that was deleted
            throw new Error('cannot read a file which is removed');
        }
        if (lodash_1.isUndefined(file.data)) {
            const data = fs_extra_1.default.readFileSync(path, 'utf-8');
            this.map.set(path, { ...file, data });
        }
        return this.getFile(path).data;
    }
    renameSync(oldPath, newPath) {
        const oldFile = this.getFile(oldPath);
        const oldFileData = this.readFileSync(oldPath);
        const oldFileStats = this.lstatSync(oldPath);
        oldFile.action = 'remove';
        const newFile = this.getFile(newPath);
        newFile.data = oldFileData;
        newFile.stats = oldFileStats;
        newFile.action = 'write';
    }
    existsSync(path) {
        if (this.map.has(path)) {
            const file = this.getFile(path);
            if (file.data && file.action !== 'remove')
                return true;
            return false;
        }
        else {
            return fs_extra_1.default.existsSync(path);
        }
    }
    ensureFileSync(path) {
        if (this.existsSync(path)) {
            return;
        }
        // File doesn't exist, create an empty one
        const file = this.getFile(path);
        file.data = '';
        // @ts-expect-error
        file.stats = new StatsLike();
        file.action = 'write';
    }
    writeChangesToDisc() {
        this.map.forEach((fileInformation, path) => {
            if (fileInformation.action === 'write') {
                fs_extra_1.default.writeFileSync(path, fileInformation.data);
            }
            else if (fileInformation.action === 'remove') {
                fs_extra_1.default.removeSync(path);
            }
        });
    }
}
exports.VirtualFileSystem = VirtualFileSystem;
class DirentLike {
    constructor(name) {
        this.isFile = () => {
            return false;
        };
        this.isDirectory = () => {
            return true;
        };
        this.isBlockDevice = () => {
            return false;
        };
        this.isCharacterDevice = () => {
            return false;
        };
        this.isSymbolicLink = () => {
            return false;
        };
        this.isFIFO = () => {
            return false;
        };
        this.isSocket = () => {
            return false;
        };
        this.name = name;
    }
}
// @ts-expect-error we only need the methods
class StatsLike {
    constructor() {
        this.isFile = () => {
            return true;
        };
        this.isDirectory = () => {
            return false;
        };
        this.isBlockDevice = () => {
            return false;
        };
        this.isCharacterDevice = () => {
            return false;
        };
        this.isSymbolicLink = () => {
            return false;
        };
        this.isFIFO = () => {
            return false;
        };
        this.isSocket = () => {
            return false;
        };
    }
}
